# -*- coding:japanese-shift-jis-dos; tab-width:4; -*-
#
# jargon-fpw.pl -- FreePWING script for The Jargon File
# $Id: jargon-fpw.pl 2.0 2004/08/25 05:52:45 satomii Exp $
#
# Copyright (C) 2000-2004 Satomi I.
#
# DO NOT CHANGE THE ENCODING OF THIS FILE.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#

use strict;
use GetOpt::Long;
use FreePWING::FPWUtils::FPWParser;
use XML::LibXML::Common;
use JargonXml;

my $VERSION = '2.0';
my $XPATH_MENU_SECTION = join(' | ', qw(
	book part preface chapter appendix sect1 sect2 glossary bibliography
));

my %ELEMENT_WRITERS = (
	# menu entries.
	book			=> \&writeMenu,
	part			=> \&writeMenuEntry,
	preface			=> \&writeMenuEntry,
	chapter			=> \&writeMenuEntry,
	appendix		=> \&writeMenuEntry,
	sect1			=> \&writeMenuEntry,
	sect2			=> \&writeMenuEntry,
	glossary		=> \&writeMenuEntry,
	glossdiv		=> \&writeMenuEntry,
	bibliography	=> \&writeMenuEntry,
	biblioentry		=> \&writeBiblioEntry,
	# text entries.
	glossentry		=> \&writeGlossEntry,
	abbrev			=> \&writeAbbrev,
	# block elements.
	blockquote		=> \&writeIndent,
	example			=> \&writeVerbatim,
	informaltable	=> \&writeTable,
	itemizedlist	=> \&writeIndent,
	listitem		=> \&writeListItem,
	literallayout	=> \&writeVerbatim,
	para			=> \&writePara,
	programlisting	=> \&writeVerbatim,
	screen			=> \&writeVerbatim,
	table			=> \&writeTable,
	title			=> \&writeTitle,
	variablelist	=> \&writeIndent,
	# inline elements.
	citetitle		=> \&writeItalic,
	emphasis		=> \&writeEmphasis,
	foreignphrase	=> \&writeItalic,
	quote			=> \&writeQuote,
	subscript		=> \&writeSubScript,
	superscript		=> \&writeSuperScript,
	# link elements.
	glossterm		=> \&writeGlossTerm,
	link			=> \&writeLink,
	ulink			=> \&writeUlink,
	# graphics.
	mediaobject		=> \&writeMediaObject,
	userchar		=> \&writeUserChar,
);

my %ENTRY_TITLE = (
	_copyright		=> 'The Jargon File, FreePWING version',
	_graphics		=> 'List of Graphics',
	glossary		=> 'Glossary',
	bibliography	=> 'Bibliography',
);

my %ENTRY_TAG = (
	_copyright		=> 'fpwCopyright',
	_graphics		=> 'fpwGraphics',
);

my $README = 'readme.txt';
my $LIST_BULLET = '* ';

my $NoGraphics = 0;
my $OutputLanguage = eval {
	use Jcode;
	if ($ENV{'OUTPUT_LANGUAGE'}) {
		$ENV{'OUTPUT_LANGUAGE'};
	} else {
		my $lang = ($ENV{'LANG'} or '');
		$lang =~ s/^[^\.]+\.//;
		$lang =~ /^(euc([\-\.]?jp)|ujis)$/i ? 'euc' : 'sjis';
	}
};

$SIG{__DIE__} = sub { &log(*STDERR, "Error: $_[0]"); exit(2); };
$SIG{__WARN__} = sub { &log(*STDERR, "Warning: $_[0]"); };

if (!GetOptions('no-graphics|t' => \$NoGraphics)
		|| 1 != int(@ARGV)) {
	die "Usage: $0 [-no-graphics] JARGON-XML-FILE\n";
}

my %FpwParser = (
	menu => undef,
	copy => undef,
	text => undef,
	head => undef,
	word => undef,
	keyword => undef,
);

my @Newline = (0, 0);
my @Indent = (0, 0);
my $Whitespace = 0;
my ($XmlParser, $XmlDoc, $MenuToc, %GlossTermID, @Graphics);

initialize_fpwparser(
	menu => \$FpwParser{menu},
	copyright => \$FpwParser{copy},
	text => \$FpwParser{text},
	heading => \$FpwParser{head},
	word2 => \$FpwParser{word},
	keyword => \$FpwParser{keyword},
);

&copyright();

$XmlParser = new JargonXml;
$XmlDoc = $XmlParser->parse_file($ARGV[0]);
&node_write($XmlDoc);
&menu_graphics() unless ($NoGraphics);

&print(sprintf("%d entries written.\n", $FpwParser{text}->entry_count()));

finalize_fpwparser(
	menu => \$FpwParser{menu},
	copyright => \$FpwParser{copy},
	text => \$FpwParser{text},
	heading => \$FpwParser{head},
	word2 => \$FpwParser{word},
	keyword => \$FpwParser{keyword},
);

exit(0);

sub log($$)
{
	my $fh = shift;
	my $msg = (eval { Jcode::convert($_[0], $OutputLanguage); } or $_[0]);
	print $fh $msg;
}

sub print($)
{
	&log(*STDOUT, $_[0]);
}

sub copyright()
{
	my $date = do {
		my ($sec, $min, $hour, $day, $mon, $year) = localtime(time());
		sprintf('%04d/%02d/%02d %02d:%02d:%02d',
				$year + 1900, $mon + 1, $day, $hour, $min, $sec);
	};
	my $fpw = $FpwParser{copy};
	my ($newline, $indent);
	local $_;

	&print("Writing copyright...\n");

	$fpw->add_entry_tag($ENTRY_TAG{_copyright})
		|| &fpw_error($fpw,
				"Unable to add copyright tag: $ENTRY_TAG{_copyright}");
	$fpw->add_keyword_start();
	$fpw->add_text($ENTRY_TITLE{_copyright})
		|| &fpw_error($fpw,
				"Unable to add text: $ENTRY_TITLE{_copyright}");
	$fpw->add_keyword_end();
	$fpw->add_newline();
	$fpw->add_text("Generated by jargon-fpw $VERSION on $date");
	$fpw->add_newline();
	$fpw->add_newline();

	$newline = 2;
	$indent = 0;

	open(IN, $README) || die "Unable to open file: $README ($!)\n";
	while (<IN>) {
		last if (/^#/);
	}
	while (<IN>) {
		last if (/^#/);
		s/[\r\n]+$//;
		s/^  //;
		if ('' eq $_) {
			if (2 != $indent) {
				$fpw->add_indent_level(2);
				$indent = 2;
			}
			while ($newline < 2) {
				$fpw->add_newline();
				$newline++;
			}
			next;
		}
		if (s/^  // && $indent < 3) {
			$fpw->add_indent_level(3);
			$indent = 3;
		}
		$fpw->add_text(' ') unless ($newline);
		$fpw->add_text($_) || &fpw_error($fpw, "Unable to add text: $_");
		if (3 == $indent) {
			$fpw->add_newline();
			$newline = 1;
		} else {
			$newline = 0;
		}
	}
	close(IN);
}

##------------------------------------------------------------------------
## FREEPWING FUNCTIONS

sub fpw_error($$)
{
	my ($fpw, $msg) = @_;
	die "$msg (" . ($fpw->error_message() or 'unknown error') . ")\n";
}

sub fpw_addText($$)
{
	my ($fpw, $text) = @_;
	my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;
	my $pos;
	local $_;

	return if ('' eq $text);

	while ('' ne $text) {
		$pos = index($text, '\\');
		if ($pos < 0) {
			$fpw->add_text($text)
				|| &fpw_error($fpw, "Unable to add text: $text");
			last;
		}
		if (0 < $pos) {
			$_ = substr($text, 0, $pos);
			$fpw->add_text($_)
				|| &fpw_error($fpw, "Unable to add text: $_");
		}
		$fpw->add_half_user_character('h005c')
			|| &fpw_error($fpw, "Unable to add half character: h005c");
		$text = substr($text, ++$pos);
	}

	$Newline[$context] = 0;
}

sub fpw_addNewline($;$)
{
	my ($fpw, $count) = @_;
	my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;

	$count = 1 if (!$count);
	while ($Newline[$context] < $count) {
		$fpw->add_newline();
		$Newline[$context]++;
	}
}

##------------------------------------------------------------------------
## NODE FUNCTIONS

sub node_error($$)
{
	my ($node, $msg) = @_;
	my $tag = eval { $node->nodeName() };

	if ($tag) {
		my @att = $node->attributes();
		while (@att) {
			$tag .= $att[0]->toString();
			shift(@att);
		}
		$tag = "<$tag>";
	} else {
		$tag = '(no name)';
	}
	die "$msg: $tag at xml line " . $node->line_number() . "\n";
}

sub node_position($)
{
	my $node = shift;
	my $pos = 0;

	while ($node = $node->previousSibling()) { $pos++; }
	$pos;
}

sub node_textLength($)
{
	my $node = shift;
	my $type = $node->nodeType();
	my $len = 0;
	my $child;
	local $_;

	if ($type == NODE_TEXT
			|| $type == NODE_ENTITY_REF) {
		$_ = decodeFromUTF8('euc-jp', $node->textContent());
		if (/^[\r\n]+[ \t]*$/) {
			return 0 if (!$node->previousSibling());
			return 0 if ('para' ne $node->parentNode()->nodeName());
		}
		s/[ \t\r\n]+/ /g;
		s/^ +// unless ($node->previousSibling());
		s/ +$// unless ($node->nextSibling());
		$len = length($_);

	} elsif ($type == NODE_ELEMENT) {
		my $name = $node->nodeName();
		if ($name eq 'userchar') {
			$_ = decodeFromUTF8('euc-jp', $node->textContent());
			if ('' ne $_) {
				$len = length($_);
			} else {
				$_ = $node->findvalue('@name');
				if ($_) {
					$_ = decodeFromUTF8('euc-jp', $_);
					$len = /^f/ ? 2 : 1;
				}
			}
		} else {
			foreach $child ($node->childNodes()) {
				$len += &node_textLength($child);
			}
			$len += 2 if ($name eq 'quote');
		}

	} elsif ($type == NODE_DOCUMENT
			|| $type == NODE_DOCUMENT_FLAG) {
		foreach $child ($node->childNodes()) {
			$len += &node_textLength($child);
		}
	}

	$len;
}

sub node_write($;$)
{
	my ($node, $fpw) = @_;
	my $type = $node->nodeType();

	if ($type == NODE_TEXT
			|| $type == NODE_ENTITY_REF) {
		local $_ = decodeFromUTF8('euc-jp', $node->textContent());
		if ($Whitespace) {
			my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;
			my @lines = split(/\r?\n/, $_);
			unless ($node->previousSibling()) {
				while (@lines && $lines[0] =~ /^$/) {
					shift(@lines);
				}
			}
			while (@lines) {
				$_ = shift(@lines);
				&fpw_addText($fpw, $_);
				if (@lines) {
					$fpw->add_newline();
					$Newline[$context] = 1;
				}
			}
		} else {
			if (/^[\r\n]+[ \t]*$/) {
				return if (!$node->previousSibling());
				return if ('para' ne $node->parentNode()->nodeName());
			}
			s/[ \t\r\n]+/ /g;
			s/^ +// unless ($node->previousSibling());
			s/ +$// unless ($node->nextSibling());
			&fpw_addText($fpw, $_);
		}

	} elsif ($type == NODE_ELEMENT) {
		my $name = $node->nodeName();
		if ($ELEMENT_WRITERS{$name}) {
			$ELEMENT_WRITERS{$name}($node, $fpw);
		} else {
			&node_parseChild($node, \&node_write, $fpw);
		}

	} elsif ($type == NODE_DOCUMENT
			|| $type == NODE_DOCUMENT_FLAG) {
		&node_parseChild($node, \&node_write, $fpw);
	}
}

sub node_parseChild($$;@)
{
	my $node = shift;
	my $handler = shift;

	# do not use XML::LibXML::NodeList->get_node(N) to enumerate the
	# elements, since somehow it may mix up the order.
	foreach my $child ($node->childNodes()) {
		&$handler($child, @_);
	}
}

##------------------------------------------------------------------------
## MENU ENTRIES

sub menu_id($)
{
	my $node = shift;
	my $id = $node->findvalue('@id') unless ('book' eq $node->nodeName());

	unless ($id) {
		# do not use line_number() which is limited to a WORD value.
		$id = sprintf("%s_%d", $node->nodeName(), &node_position($node) + 1);
		$node->addChild($XmlDoc->createAttribute('id', $id));
	}
	$id;
}

sub menu_addToc($;$)
{
	my ($node, $indent) = @_;
	my (@nodes, @title, $bull, $name, $id);
	my $fpw = $FpwParser{menu};

	$name = $node->nodeName();
	if ($name eq 'glossary') {
		return if ($indent);
		@nodes = $node->findnodes('glossdiv');
	} elsif ($name eq 'glossdiv') {
		return if ($indent);
		@nodes = $node->findnodes('glossentry');
	} else {
		@nodes = $node->findnodes($XPATH_MENU_SECTION);
		$bull = $LIST_BULLET;
	}
	return if (!@nodes);

	if ($indent) {
		$fpw->add_indent_level(++$Indent[0]);
	} elsif ($name !~ /^gloss/) {
		$fpw->add_font_start('bold');
		&fpw_addText($fpw, 'Table of Contents');
		$fpw->add_font_end();
		&fpw_addNewline($fpw, 2);
	}

	foreach my $sect (@nodes) {
		$name = $sect->nodeName();
		$id = &menu_id($sect);
		$fpw->add_text($bull) if ($bull);
		$fpw->add_reference_start();

		if ($name eq 'appendix' && $id =~ /^appendix([a-z])$/) {
			&fpw_addText($fpw, 'Appendix ' . uc($1) . ': ');
		}
		@title = $sect->findnodes(($name eq 'glossentry') ?
				'glossterm' : 'title');
		if (@title) {
			foreach my $child (@title) {
				&node_parseChild($child, \&node_write, $fpw);
			}
		} else {
			local $_ = ($ENTRY_TITLE{$id} or $ENTRY_TITLE{$name})
				|| warn "Missing title: <$name id=\"$id\">\n";
			&fpw_addText($fpw, $_);
		}

		$fpw->add_reference_end($id);
		&fpw_addNewline($fpw, 1);
		&menu_addToc($sect, $indent + 1);
	}

	if ($indent) {
		$fpw->add_indent_level(--$Indent[0]);
	} else {
		&fpw_addNewline($fpw, 2);
	}
}

sub menu_graphics()
{
	my $fpw = $FpwParser{menu};
	my ($node, $parent, $file, $id);
	local $_;

	&print("Making graphics list...\n");

	$fpw->new_context() || &fpw_error($fpw, "Unable to add graphics menu");
	$fpw->add_entry_tag($ENTRY_TAG{_graphics})
		|| &fpw_error($fpw, "Unable to add menu tag: $ENTRY_TAG{_graphics}");

	foreach $node (@Graphics) {
		$_ = decodeFromUTF8('euc-jp', $node->findvalue('@fileref'));
		s/^.*\/([^\/]+)$/$1/;
		$file = $_;
		$file =~ s/\./_/g;

		$fpw->new_entry() || &fpw_error($fpw, "Unable to add menu entry");
		$fpw->add_keyword_start();
		&fpw_addText($fpw, $_);
		$fpw->add_keyword_end();
		&fpw_addNewline($fpw, 1);

		&fpw_addText($fpw, '>> ');
		$parent = $node->findnodes('ancestor::glossentry');

		if ($parent) {
			$parent = $parent->item(0);
			$id = decodeFromUTF8('euc-jp', $parent->findvalue('@id'));
			$fpw->add_reference_start();
			&node_parseChild($parent->findnodes('glossterm')->item(0),
					\&node_write, $fpw);
			$fpw->add_reference_end($id)
				|| &fpw_error($fpw, "Unable to make xref: $id");

		} else {
			my $regexp = $XPATH_MENU_SECTION;
			$regexp =~ s/ //g;
			$parent = $node;

			while ($parent = $parent->parentNode()) {
				next unless (NODE_ELEMENT == $parent->nodeType());
				next unless ($parent->nodeName() =~ /^($regexp)$/);

				$id = $parent->findvalue('@id') || next;
				$id = decodeFromUTF8('euc-jp', $id);
				$fpw->add_reference_start();

				my $title = $parent->findnodes('title');
				if ($title) {
					&node_parseChild($title->item(0), \&node_write, $fpw);
				} else {
					&fpw_addText($fpw,
							$ENTRY_TITLE{$id} or
							$ENTRY_TITLE{$parent->nodeName()});
				}

				$fpw->add_reference_end($id)
					|| &fpw_error($fpw, "Unable to make xref: $id");
				last;
			}

			unless ($parent) {
				warn "Parent section could not be found: $file\n";
				&fpw_addText($fpw, '(entry unknown)');
			}
		}

		&fpw_addNewline($fpw, 1);
		$fpw->add_color_graphic_start($file)
			|| &fpw_error($fpw, "Unable to add graphic: $file");
		$fpw->add_color_graphic_end();
		$fpw->add_newline();
		$fpw->add_newline();
	}
}

sub writeMenu($;$)
{
	my $node = shift;
	my $fpw = $FpwParser{menu};		# ignroe $_[1].
	my $id = &menu_id($node);
	local $_ = '';

	# link to the copyright entry.
	$fpw->add_reference_start();
	&fpw_addText($fpw, $ENTRY_TITLE{_copyright});
	$fpw->add_reference_end($ENTRY_TAG{_copyright})
		|| &fpw_error($fpw, "Unable to make xref: $ENTRY_TAG{_copyright}");
	&fpw_addNewline($fpw, 1);

	# link to the jargon documentation.
	$fpw->add_reference_start();
	foreach my $child ($node->findnodes('title | subtitle')) {
		if ($_) { $fpw->add_text($_); }
		else { $_ = ' '; };
		&node_parseChild($child, \&node_write, $fpw);
	}
	$fpw->add_reference_end($id)
		|| &fpw_error($fpw, "Unable to make xref: $id");
	&fpw_addNewline($fpw, 1);

	# link to the list of graphics.
	unless ($NoGraphics) {
		$fpw->add_reference_start();
		&fpw_addText($fpw, $ENTRY_TITLE{_graphics});
		$fpw->add_reference_end($ENTRY_TAG{_graphics})
			|| &fpw_error($fpw, "Unable to make xref: $ENTRY_TAG{graphics}");
		&fpw_addNewline($fpw, 1);
	}

	&fpw_addNewline($fpw, 2);
	&writeMenuEntry($node);
}

sub writeMenuEntry($;$)
{
	my $node = shift;
	my $fpw = $FpwParser{menu};		# ignore $_[1].
	my $name = $node->nodeName();
		&fpw_addNewline($fpw, 1);
	my $id = &menu_id($node);
	my (@nodes, $child);
	local $_;

	if ($MenuToc) {
		&menu_addToc($MenuToc);
		undef $MenuToc;
	}

	if ($name =~ /^sect/ || $name eq 'glossdiv') {
		$fpw->new_entry()
			|| &fpw_error($fpw, "Unable to add menu entry: $id");
	} else {
		$fpw->new_context()
			|| &fpw_error($fpw, "Unable to add menu context: $id");
	}
	$fpw->add_entry_tag($id)
		|| &fpw_error($fpw, "Unable to add menu tag: $id");

	@nodes = $node->findnodes('title');
	if (@nodes) {
		$_ = '';
		foreach $child (@nodes) {
			$_ .= ' ' if ('' ne $_);
			$_ .= $child->textContent();
		}
		s/[ \t\r\n]+/ /g;
	} else {
		# this node may not contain the title.
		my $name = $node->nodeName();
		$_ = ($ENTRY_TITLE{$id} or $ENTRY_TITLE{$name})
			|| warn "Missing title: <$name id=\"$id\">\n";
	}

	&print("Adding menu entry: $_\n");

	# write title.
	if ($name eq 'appendix' && $id =~ /^appendix([a-z])$/) {
		$fpw->add_keyword_start();
		&fpw_addText($fpw, 'Appendix ' . uc($1) . ': ');
		$fpw->add_keyword_end();
	}
	if (@nodes) {
		foreach $child (@nodes) {
			&writeKeyword($child, $fpw);
			&fpw_addNewline($fpw, 1);
		}
	} else {
		$fpw->add_keyword_start();
		&fpw_addText($fpw, $_);
		$fpw->add_keyword_end();
		&fpw_addNewline($fpw, 1);
	}
	foreach $child ($node->findnodes('subtitle')) {
		&node_parseChild($child, \&node_write, $fpw);
		&fpw_addNewline($fpw, 1);
	}

	&fpw_addNewline($fpw, 2);
	$fpw->add_indent_level(2);
	$Indent[0] = 2;
	$MenuToc = $node;

	# parse children other than title.
	foreach $child ($node->findnodes('*')) {
		my $name = $child->nodeName();
		if ($name !~ /^(sub)?title$/) {
			&node_write($child, $fpw);
		}
	}

	&fpw_addNewline($fpw, 2);
}

sub writeBiblioEntry($$)
{
	my ($node, $fpw) = @_;
	my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;
	my $id = $node->findvalue('@id');
	my ($child, $token, $i, $j);

	&fpw_addNewline($fpw, 2);
	if ($id) {
		$id = decodeFromUTF8('euc-jp', $id);
		$fpw->add_entry_tag($id)
			|| &fpw_error($fpw, "Unable to add bibliography tag: $id");
	}

	foreach $child ($node->findnodes('title')) {
		&writeBold($child, $fpw);
		&fpw_addNewline($fpw, 1);
	}

	$i = 0;
	foreach my $child ($node->findnodes('author | authorgroup/author')) {
		&fpw_addText($fpw, ' and ') if ($i);
		$j = 0;
		foreach $token ($child->findnodes('*')) {
			&fpw_addText($fpw, ' ') if ($j);
			&node_parseChild($token, \&node_write, $fpw);
			$j++;
		}
		$i++;
	}
	if ($i) {
		&fpw_addText($fpw, '.');
		$i = 0;
	}

	foreach $child ($node->findnodes('*')) {
		my $name = $child->nodeName();
		if ($name =~ /^(title|author(group)?)$/) {
			# nop.
		} elsif ($name eq 'abstract') {
			$fpw->add_indent_level(++$Indent[$context]);
			&node_parseChild($child, \&node_write, $fpw);
			$fpw->add_indent_level(--$Indent[$context]);
			$i = 0;
		} else {
			&fpw_addText($fpw, ' ') if ($i);
			if ($name eq 'copyright') {
				&fpw_addText($fpw, 'Copyright (C) ');
			}
			$j = 0;
			foreach $token ($child->findnodes('descendant::text()')) {
				local $_ = decodeFromUTF8('euc-jp', $token->textContent());
				next if (/^[\r\n]+[ \t]*$/);
				s/[ \t\r\n]+/ /g;
				s/^ +//;
				s/ +$//;
				$_ = " $_" if ($j);
				&fpw_addText($fpw, $_);
				$j++;
			}
			&fpw_addText($fpw, '.');
		}
		$i++;
	}
}

##------------------------------------------------------------------------
## TEXT ENTRIES

sub writeGlossEntry($;$)
{
	my $node = shift;
	my $fpwtext = $FpwParser{text};		# ignore $_[1].
	my $fpwhead = $FpwParser{head};
	my $id = $node->findvalue('@id');
	my ($child, $hpos, $tpos, @words);
	local $_;

	&node_error($node, "Missing id attribute") unless ($id);
	$id = decodeFromUTF8('euc-jp', $id);

	$fpwhead->new_entry()
		|| &fpw_error($fpwhead, "Unable to add heading entry: $id");
	$fpwtext->new_entry()
		|| &fpw_error($fpwtext, "Unable to add text entry: $id");
	$fpwtext->add_entry_tag($id)
		|| &fpw_error($fpwtext, "Unable to add entry tag: $id");

	# retrieve the glossterm node.
	$child = do {
		my $list = $node->findnodes('glossterm[1]')
			|| &node_error($node, "Missing glossterm");
		$list->item(0);
	};

	$_ = $child->textContent();
	s/[ \t\r\n]+/ /g;
	$GlossTermID{$_} = $id;
	$_ = decodeFromUTF8('euc-jp', $_);
	&print("Adding text entry: $_\n");

	# write the heading.
	$fpwhead->add_text($_)
		|| &fpw_error($fpwhead, "Unable to write heading: $_");
	&writeKeyword($child, $fpwtext);
	&fpw_addNewline($fpwtext, 1);

	# register search words for this entry.
	$hpos = $fpwhead->entry_position();
	$tpos = $fpwtext->entry_position();
	$FpwParser{word}->add_entry($_, $hpos, $tpos)
		|| &fpw_error($FpwParser{word}, "Unable to add search word: $_");

	@words = split(/[ \(\)]+/, $_);
	foreach my $word (@words) {
		$word =~ s/[\.,!\?]//g;
		next if ('' eq $word || $_ eq $word);
		$FpwParser{keyword}->add_entry($word, $hpos, $tpos)
			|| &fpw_error($FpwParser{keyword},
					"Unable to add conditional keyword: $word");
	}

	# write pronunciation and grammar information if any.
	$_ = '';
	foreach my $abbrev ($node->findnodes('abbrev')) {
		foreach $child ($abbrev->findnodes('*')) {
			if ($_) { $fpwtext->add_text($_); }
			else { $_ = ', '; }
			&node_parseChild($child, \&node_write, $fpwtext);
		}
	}

	# write definitions.
	&fpw_addNewline($fpwtext, 2);
	$fpwtext->add_indent_level(2);
	$Indent[1] = 2;

	# glossdef should not contain #PCDATA.
	foreach $child ($node->findnodes('glossdef/*')) {
		&node_write($child, $fpwtext);
	}

	&fpw_addNewline($fpwtext, 2);
}

sub writeAbbrev($$)
{
	my ($node, $fpw) = @_;

	&node_parseChild($node, \&node_write, $fpw);
	if ('glossentry' eq $node->parentNode()->nodeName()) {
		&fpw_addNewline($fpw, 1);
	}
}

##------------------------------------------------------------------------
## BLOCK ELEMENTS

sub writeTitle($$)
{
	my ($node, $fpw) = @_;

	&fpw_addNewline($fpw, 2);
	&writeBold($node, $fpw);
	&fpw_addNewline($fpw, 2);
}

sub writePara($$)
{
	my ($node, $fpw) = @_;

	&fpw_addNewline($fpw, 2);
	&node_parseChild($node, \&node_write, $fpw);
	&fpw_addNewline($fpw, 2);
}

sub writeIndent($$)
{
	my ($node, $fpw) = @_;
	my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;

	&fpw_addNewline($fpw, 2);
	$fpw->add_indent_level(++$Indent[$context]);
	&node_parseChild($node, \&node_write, $fpw);
	&fpw_addNewline($fpw, 1);
	$fpw->add_indent_level(--$Indent[$context]);
	&fpw_addNewline($fpw, 2);
}

sub writeVerbatim($$)
{
	die "Unexpected error" if ($Whitespace < 0);
	$Whitespace++;
	&writeIndent(@_);
	$Whitespace--;
}

sub writeListItem($$)
{
	my ($node, $fpw) = @_;
	my $parent = $node->parentNode()->nodeName();
	my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;

	&fpw_addNewline($fpw, 1);

	if ($parent eq 'itemizedlist') {
		$fpw->add_text($LIST_BULLET);
		$Newline[$context] = 2;		# dummy.
		&node_parseChild($node, \&node_write, $fpw);

	} elsif ($parent eq 'varlistentry') {
		$fpw->add_indent_level(++$Indent[$context]);
		&node_parseChild($node, \&node_write, $fpw);
		$fpw->add_indent_level(--$Indent[$context]);
	}
}

sub writeTable($$)
{
	my ($node, $fpw) = @_;
	my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;
	my $cols = $node->findvalue('tgroup/@cols');
	my (@maxlen, $len, $row, $col, $i);

	&fpw_addNewline($fpw, 2);
	$fpw->add_indent_level(++$Indent[$context]);

	foreach $row ($node->findnodes('tgroup/*/row')) {
		$i = 0;
		foreach $col ($row->findnodes('entry')) {
			$len = &node_textLength($col) + 2;
			$maxlen[$i] = $len if ($maxlen[$i] < $len);
			$i++;
		}
	}

	$len = 0;
	for ($i = 0; $i <= $#maxlen; $i++) {
		$len += $maxlen[$i];
	}

	if ($len < 75) {
		foreach $row ($node->findnodes('tgroup/*/row')) {
			$i = 0;
			foreach $col ($row->findnodes('entry')) {
				&node_parseChild($col, \&node_write, $fpw);
				if (0 < $col->findvalue('count(following-sibling::entry)')) {
					$len = $maxlen[$i] - &node_textLength($col);
					$fpw->add_text(' ' x $len) if (0 < $len);
				}
				$i++;
			}
			&fpw_addNewline($fpw, 1);
		}

	} elsif (!$node->findvalue('count(tgroup/thead)')
			&& 2 == $node->findvalue('tgroup/@cols')) {
		foreach $row ($node->findnodes('tgroup/*/row')) {
			$i = 0;
			foreach $col ($row->findnodes('entry')) {
				if ($i) {
					$fpw->add_indent_level(++$Indent[$context]);
					&node_parseChild($col, \&node_write, $fpw);
					$fpw->add_indent_level(--$Indent[$context]);
				} else {
					&node_parseChild($col, \&node_write, $fpw);
				}
				&fpw_addNewline($fpw, 1);
				$i++;
			}
			&fpw_addNewline($fpw, 2);
		}

	} else {
		my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;
		my @header;

		$i = 0;
		foreach $col ($node->findnodes('tgroup/thead/row/entry')) {
			@header[$i++] = $col;
		}
		foreach $row ($node->findnodes('tgroup/tbody/row')) {
			$i = 0;
			foreach $col ($row->findnodes('entry')) {
				if ($header[$i]) {
					&node_parseChild($header[$i], \&node_write, $fpw);
					$fpw->add_text(': ');
					$Newline[$context] = 2;		# dummy.
				}
				&node_parseChild($col, \&node_write, $fpw);
				&fpw_addNewline($fpw, 1);
				$i++;
			}
			&fpw_addNewline($fpw, 2);
		}
	}

	$fpw->add_indent_level(--$Indent[$context]);
	&fpw_addNewline($fpw, 2);
}

##------------------------------------------------------------------------
## INLINE ELEMENTS

sub writeKeyword($$)
{
	my ($node, $fpw) = @_;

	$fpw->add_keyword_start();
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_keyword_end();
}

sub writeBold($$)
{
	my ($node, $fpw) = @_;

	$fpw->add_font_start('bold');
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_font_end();
}

sub writeItalic($$)
{
	my ($node, $fpw) = @_;

	$fpw->add_font_start('italic')
		|| &fpw_error($fpw, "Unable to change font to italic: "
				. $node->textContent()
				. " at xml line " . $node->line_number());
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_font_end();
}

sub writeEmphasis($$)
{
	my ($node, $fpw) = @_;
	my $role = ($node->findvalue('@role') or $node->findvalue('remap'));

	if ($role eq 'pronunciation'
			|| $role eq 'grammar'
			|| $role eq 'fraktur') {
		&node_parseChild($node, \&node_write, $fpw);
	} else {
		&writeItalic(@_);
	}
}

sub writeQuote($$)
{
	my ($node, $fpw) = @_;

	$fpw->add_text('"');
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_text('"');
}

sub writeSubScript($$)
{
	my ($node, $fpw) = @_;

	$fpw->add_subscript_start();
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_subscript_end();
}

sub writeSuperScript($$)
{
	my ($node, $fpw) = @_;

	$fpw->add_superscript_start();
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_superscript_end();
}

##------------------------------------------------------------------------
## LINK ELEMENTS

sub writeGlossTerm($$)
{
	my ($node, $fpw) = @_;
	my $id = $node->findvalue('@linkend');

	# glossterm should not be nested.
	if ($node->findnodes('ancestor::glossterm')) {
		warn "Invalid xml: glossterm should not be nested.\n";
		&node_parseChild($node, \&node_write, $fpw);
		return;
	}

	unless ($id) {
		# if the target term contains an un-safe character such as
		# apostrophe, it should be represented as an entity, which however
		# causes the XPath to match nothing. so we first obtain all
		# glossentry nodes and parse them in a loop to find a match. any
		# easier workaround?
		my $term = $node->textContent();
		$term =~ s/[ \t\r\n]+/ /g;

		if (defined($GlossTermID{$term})) {
			$id = $GlossTermID{$term};
		} else {
			foreach my $node ($XmlDoc->findnodes('//glossentry/glossterm')) {
				local $_ = $node->textContent();
				s/[ \t\r\n]+/ /g;
				if ($_ eq $term) {
					$id = $node->parentNode()->findvalue('@id')
						|| &node_error($node, "Missing id attribute");
					last;
				}
			}
			unless ($id) {
				warn "Xref linkend not found: $term\n";
				return;
			}
			$id = decodeFromUTF8('euc-jp', $id);
			$GlossTermID{$term} = $id;
		}
	}

	$fpw->add_reference_start();
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_reference_end($id)
		|| &fpw_error($fpw, "Unable to make xref: $id");
}

sub writeLink($$)
{
	my ($node, $fpw) = @_;
	my $tag = decodeFromUTF8('euc-jp', $node->findvalue('@linkend'));

	$fpw->add_reference_start();
	&node_parseChild($node, \&node_write, $fpw);
	$fpw->add_reference_end($tag)
		|| &fpw_error($fpw, "Unable to make xref: $tag");
}

sub writeUlink($$)
{
	my ($node, $fpw) = @_;
	my $url = $node->findvalue('@url');

	&node_parseChild($node, \&node_write, $fpw);
	if ($url && $url ne $node->textContent()) {
		$url = decodeFromUTF8('euc-jp', $url);
		&fpw_addText($fpw, " <$url>");
	}
}

##------------------------------------------------------------------------
## GRAPHICS

sub writeMediaObject($$)
{
	my ($node, $fpw) = @_;
	return if ($NoGraphics);

	my @image = $node->findnodes('imageobject/imagedata[@fileref]');
	return if (!@image);

	my $context = ($fpw == $FpwParser{menu}) ? 0 : 1;
	my $id = $node->findvalue('@id');
	my ($child, $file);
	local $_;

	&fpw_addNewline($fpw, 2);
	$fpw->add_indent_level(++$Indent[$context]);
	if ($id) {
		$fpw->add_entry_tag($id)
			|| &fpw_error($fpw, "Unable to add graphic tag: $id");
	}

	foreach $child (@image) {
		$_ = decodeFromUTF8('euc-jp', $child->findvalue('@fileref'));
		s/^.*\/([^\/]+)$/$1/;
		$file = $_;
		$file =~ s/\./_/g;
		$fpw->add_color_graphic_start($file)
			|| &fpw_error($fpw, "Unable to add graphic: $file");
		&fpw_addText($fpw, $_);
		$fpw->add_color_graphic_end();
		&fpw_addNewline($fpw, 1);
		push(@Graphics, $child);
	}

	$Newline[$context] = 2;
	foreach $child ($node->findnodes('caption/*')) {
		if ('()' ne $child->textContent() || $child->findnodes('*')) {
			&node_write($child, $fpw);
		}
	}

	$fpw->add_indent_level(--$Indent[$context]);
}

sub writeUserChar($$)
{
	my ($node, $fpw) = @_;
	local $_;

	$_ = decodeFromUTF8('euc-jp', $node->textContent());
	if ('' ne $_ && !$node->findvalue('count(ancestor::screen)')) {
		&fpw_addText($fpw, $_);
		return;
	}

	$_ = $node->findvalue('@name')
		|| &node_error($node, "Missing name attribute");
	$_ = decodeFromUTF8('euc-jp', $_);
	if (/^f/) {
		$fpw->add_full_user_character($_)
			|| &fpw_error($fpw, "Unable to add user character: $_");
	} else {
		$fpw->add_half_user_character($_)
			|| &fpw_error($fpw, "Unable to add user character: $_");
	}
	$Newline[($fpw == $FpwParser{menu}) ? 0 : 1] = 0;
}
